@@ -19,11 +19,11 @@ libLLVM_handle = C_NULL
 libLLVM_path = ""
 
 if Sys.iswindows()
-    const libLLVM = "libLLVM-13jl.dll"
+    const libLLVM = "libLLVM-14jl.dll"
 elseif Sys.isapple()
     const libLLVM = "@rpath/libLLVM.dylib"
 else
-    const libLLVM = "libLLVM.so"
+    const libLLVM = "libLLVM-15.so"
 end
 
 function __init__()
